


class CompoundMedia:
    """
    Class to represent a media that is built with
    different medias at the same time. Useful when
    we are creating a scene on a track but we want
    to make it be together and dependant. For 
    example we have a main video and we create an
    overlay and an animation that are displayed at
    the same time.
    """
    pass

# TODO: Create it, it is a bit complex because
# it has to imitate the Timeline